'''Helper functions for survey and create_survey
are stored here.
'''

from config import *
from random import randrange
import os 

######## HELPER FUNCTIONS FOR AUTH #########

def listdir(dir_name):
    '''Return os.listdir, but create the directory if it doesn't exist.
    '''
    try:
        return os.listdir(dir_name)
    except(OSError):
        os.makedirs(dir_name)
        return os.listdir(dir_name)

####### HELPER FUNCTIONS FOR WRITE_SURVEY #######

def debug_print(text):
    if DEBUGGING:
        print 'Content-type: text/html\n\n' + text

def get_code(number):
    '''Returns the question code for the given number.  This is for
       the initialization function of the survey class and the
       creation of the input dictionaries for the survey class (in
       input_survey).
    '''
    return 'q%03d' % number

def get_key(code, key):
    '''Return a key for the dictionary input to the survey class.
       The code is the question code returned by helpers.get_code,
       and the key is the part of the question that is being
       encoded (e.g. text, options, likert, etc.).
    '''
    if not code:
        return key
    return code+'-'+key

def get_value(dict, code, key, return_blank=0):
    '''Return the value in the dict keyed with the question code
       and the key given.  Used in the survey class.
    '''
    if return_blank:
        return dict.get(get_key(code, key), '')
    else:
        return dict[get_key(code, key)]

def likert_name(code, index):
    '''Returns an option name to be used in likert scale questions
       in the survey class.
    '''
    return code + '_%02d' % index

####### HELPER FUNCTIONS FOR PROCESS_SURVEY #######

def remove_strings(string, dict, reverse=0, keys=[]):
    if not keys:
        keys = dict.keys()
    if reverse:
        for key in keys:
            if dict[key]:
                string = string.replace(dict[key], key)
    else:
        for key in keys:
            string = string.replace(key, dict[key])
    return string

def html_escape(string, reverse=0):
    '''Return a string that will be displayed as-is in
       HTML browsers and that is safe for HTML forms
       (with <, >, ", \n characters 'escaped', so that
       they are just displayed rather than affecting the
       HTML of the page).
       If reverse is specified, undo the HTML escaping.
    '''
    return remove_strings(string, html_dict, reverse, html_dict_keys)

def htmlize(string, reverse=0):
    '''Return a string that will look the in a web browser
       as it does in a textfile.  Replace newlines with <br>.
    '''
    return remove_strings(string, htmlize_dict, reverse, htmlize_dict_keys)

def form_escape(text, reverse=0):
    '''Return a string that can be used as a value in a form
       element, with spaces converted to underscores and
       punctuation removed.
    '''
    return remove_strings(text, punctuation_dict, reverse)

def csv_escape(text):
    '''
    '''
    return remove_strings(text, csv_dict, reverse=0)   

def get_survey_value(dict, att, return_empty=0, multivalue=0):
    '''Retrieve the value of the attribute specified in the dictionary given.
    '''
    if not return_empty:
        to_return = dict[att]
    else:
        to_return = dict.get(att, [''])
    if not multivalue and type(to_return) == type([]):
        return to_return[0]
    else:
        return to_return
    
def has_value(dict, att):
    '''Return 1 if the dictionary contains the given attribute, 0 otherwise.
    '''
    if att in dict:
        return 1
    else:
        return 0

def get_pagenum(dict):
    '''Retrieve the survey page number (as an int) from the dictionary.
    '''
    return int(get_survey_value(dict, PAGENUM))

def get_header(survey_id=''):
    '''
    '''
    file = open('header.txt')
    text = file.read() % {'survey_tool_name': SURVEY_TOOL_NAME, \
                          'survey_id': survey_id, \
                          'base_url': BASE_URL+survey_id}
    file.close()
    return text

def get_footer(survey_id=''):
    '''
    '''
    file = open('footer.txt')
    text = file.read() % {'survey_tool_name': SURVEY_TOOL_NAME_LINK, \
                          'survey_id': survey_id}
    file.close()
    return text

def get_survey_url(s_id):
    return BASE_URL + 'index.cgi?' + SURVEY_ID + '=' + s_id

def get_survey_dir(survey_id, data=1, include_path=1):
    '''Return the path to the directory for the specified survey (or its
       data directory, if specified).
    '''
    path = survey_id+'/'
    if include_path:
        path = SURVEY_PATH + '/' + path
    if data:
        path += DATA_DIR+'/'
    return path

def _substring_in_list(sub, list):
    inlist = 0
    for item in list:
        if item.find(sub) > -1:
            inlist = 1
    return inlist

def make_id(survey_id):
    '''Generate a random ID string from a hex number.
    '''
    files = SURVEY_PATH + '/'
    if survey_id:
        files = os.listdir(get_survey_dir(survey_id))
    id = str(hex(randrange(0, 1000000000)))[2:]
    while _substring_in_list(id, files):
        id = str(hex(randrange(0, 1000000000)))[2:]
    return id

def list_strip(list):
    '''Return a copy of the list with empty items removed
       and whitespace stripped from other items in the list.
       Used in input_survey.
    '''
    to_return = []
    for line in list:
        if line:
            to_return.append(line.strip())
    return to_return

def htaccess(survey_id, user_id, append=0):
    survey_dir = get_survey_dir(survey_id, data=0, include_path=0)
    htaccess_text = ''
    if append:
        file_access = 'a'
    else:
        file_access = 'w'
        htaccess_text += 'AuthUserFile ' + BASE_PATH + survey_dir + '''.htpasswd
AuthGroupFile /dev/null
AuthName EnterPassword
AuthType Basic

'''
    htaccess_text += 'require user ' + user_id
    file = open(survey_dir + '.htaccess', file_access)
    file.write(htaccess_text)
    file.close()    

def htpasswd_create(user_id, password):
    output = os.system((HTPASSWD_LOCATION+HTPASSWD_CREATE+USERS_DIR+user_id+'.htpasswd '+user_id+' '+password))
    if debugging:
        print 'htpasswd output (0 means success):', output, '<p>\n\n'

def htpasswd(survey_id, user_id, append=0):
    if append:
        file_access = 'a'
    else:
        file_access = 'w'
    survey_dir = get_survey_dir(survey_id, data=0, include_path=0)
    file = open(USERS_DIR+user_id+'.htpasswd')
    htpasswd_string = file.read()
    if not htpasswd_string.endswith('\n'):
        htpasswd_string += '\n'
    file.close()
    file = open(survey_dir+'.htpasswd', file_access)
    file.write(htpasswd_string)
    file.close()
    # output = os.system((HTPASSWD_LOCATION+file_access+survey_dir+'.htpasswd '+user_id+' '+password))
    # if debugging:
    #    print 'htpasswd output (0 means success):', output, '<p>\n\n'

def secure_survey(survey_id, user_id, append=0):
    '''Requires survey directory to already exist.
    '''
    htaccess(survey_id, user_id, append)
    htpasswd(survey_id, user_id, append)

def forward(destination):
    """Forward to the specific destination, relative to the BASE_URL.  If 
       arguments are given after a cgi script (e.g. foo.cgi?arg=val), spaces 
       should be represented as plusses, name-value pairs should be
       separated by ampersands (&), and there should be no '<' or '>'
       characters.
    """
    return """Content-Type: text/html

<html>
<head>
  <meta http-equiv="Refresh" content="0; URL=%(base)s%(dest)s">
</head>
<body>
  <font color="white">
  Forwarding you to <a href="%(base)s%(dest)s"><font color="white">%(base)s%(dest)s</font></a>
  </font>
</body>
</html>
""" % {'base': BASE_URL, 'dest': destination}
