'''
EASY SURVEY by Morgan Ames (http://www.ocf.berkeley.edu/~morganya)
Easily create, edit, distribute, and output data from online surveys.

Installation instructions:
Easy Survey is simple to install if your web server allows execution of
.cgi scripts in any directory.

1. Unzip Easy Survey into a location of your choice.  It will unzip
   into a directory called "survey", but feel free to change this.
2. Change the main directory and files with .cgi extensions to be
   executable (in the main survey directory, type "chmod 755 *.cgi").
   All other files should be read-only by default.
3. Update BASE_URL and BASE_PATH in config.py to the location of
   the survey directory.

This program was written or based on one written in 2004 by Morgan Ames.
http://www.ocf.berkeley.edu/~morganya

This program is free software; you can redistribute it and/or modify
it under the terms of the Creative Commons Attribution-NonCommercial 
License. 

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE (as described by
the GNU General Public License at http://www.gnu.org/licenses/gpl.html). 

To view a copy of the Creative Commons Attribution-NonCommercial license, 
visit http://creativecommons.org/licenses/by-nc/2.0/ or send a letter to 
Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.

UPCOMING FEATURES:

In v. 0.18:
- optional survey-taker logins (for a restricted survey)
- email notification for survey-takers
- tracking of survey-takers (for restricted and non-restricted surveys)
- data encryption (for CPHS)
- ability to transfer surveys to other users

In v. 0.2:
- "super-user" or manager account
- setup from browser
- ability to delete users
- implement or augment error handling functions
- optional status bar at the bottom of the page (page x of y, or question k of n)
- ability to add new variables such as %PAGE% that can be used in survey creation
- prevent response errors due to modifications to survey after responses are
  made - potential options:
   - LOCK: "lock" survey, restricting changes that can be made to just rewording
   - VERSIONING: save all survey versions and keep track of responses to each
     version (still problem of reporting results from different versions, though
     question text is saved with answer)
   - only allow web-based edits after responses have been made, and assign
     a code to each question
- ability to share (revoke?) survey editing privileges with other users (??)

In v. 0.3:
- recursive questions for text-based input (is this actually needed?)
- form-based survey input (depends on success of text-based input)
- export to SPSS, HTML
- implement conditionals (different survey content based on previous responses)
- implement required questions
- database integration?
'''

import os, pickle, time
from helpers import *

class question:
    atts = {}
    
    def __init__(self, dict):
        '''Types of questions:
           vertical radio
           horizontal radio (likert)
           vertical checkbox
           horizontal checkbox
           selection
           one-line text
           password
           textarea
           pagebreak
           page title
           page text
           reset
        Question attributes:
           type (required), code (required),
           text, options, likert, required, condition
        '''
        self.atts = dict

    def _hidden_field(self, index=0):
        if index:
            name = self.get_name(likert_index=index)
        else:
            name = get_key(self.get_name(), QTEXT)
        return '<input type="hidden" name="' + name + \
               '" value="' + html_escape(self.get(QTEXT)) + '">\n'

    def get_name(self, checkbox_index=0, likert_index=0):
        '''Return the code name used in the question
           form elements.
        '''
        q_name = self.get(QCODE, 'test') # TODO fix? this alternate was added so we can preview questions, but could break surveys that somehow don't assign question codes.
        type = self.get_type()
        if type == H_CHECKBOX or type == H_RADIO:
            q_name = likert_name(q_name, likert_index)
        if type == V_CHECKBOX or type == H_CHECKBOX:
            q_name += '_%02d' % (checkbox_index+1)
        return q_name

    def get_names(self):
        '''Return a list of all the codes used in the
           question form elements.
        '''
        names = []
        type = self.get_type()
        if type == H_CHECKBOX or type == H_RADIO or \
           type == V_CHECKBOX:
            option_index = 0
            for option in self.get(QOPTIONS):
                if type == H_CHECKBOX:
                    likert_index = 0
                    for likert in self.get(QLIKERT):
                        names.append(self.get_name(likert_index, option_index))
                        likert_index += 1
                else:
                    names.append(self.get_name(option_index))
                if self._isother(option):
                    names.append(self._other_name())
                option_index += 1
        else:
            names.append(self.get_name())
        return names
            

    def _question_beginning(self):
        return '<li><span class="questiontext">' + \
               self.get(QTEXT) + '</span>\n'

    def _isother(self, option):
        return option.lower().strip().strip(':') == 'other'

    def _other_name(self):
        return self.get_name() + OTHER_FIELD_NAME

    def _other_field(self, option, style):
        string = ''
        if self._isother(option):
            if not option.endswith(':'):
                string += ':'
            if style == TEXTINPUT:
                string += ' <input type="text" name="' + \
                          self._other_name() + \
                          '" size="' + str(OTHER_TEXTINPUT_SIZE) + \
                          '">\n'
            elif style == TEXTAREA:
                string += '<br><textarea name="' + \
                          self.get_name() + OTHER_FIELD_NAME + \
                          '" cols="' + str(OTHER_TEXTAREA_WIDTH) + \
                          '" rows="' + str(OTHER_TEXTAREA_HEIGHT) + \
                          '"></textarea>\n'
        return string

    def _vertical_element(self, type):
        string = '<table class="v_element">\n'
        checkbox_index = 0
        for option in self.get(QOPTIONS):
            string += '<tr>\n  <td class="v_element"><input type="'+type+'" name="' + \
                      self.get_name(checkbox_index) + \
                      '" value="' + form_escape(option) + \
                      '"></td>\n  <td class="v_element">' + option + self._other_field(option, TEXTINPUT) + \
                      '</td>\n</tr>\n'
            checkbox_index += 1
        return string + '</table>'

    def _horizontal_element(self, type):
        string = '<table>\n<tr>\n  <td></td>\n'
        for value in self.get(QLIKERT):
            string += '  <td class="likert">' + value + '</td>\n'
        string += '</tr>\n'
        likert_index = 1
        for option in self.get(QOPTIONS):
            if likert_index%2:
                parity = 'odd'
            else:
                parity = 'even'
            string += '<tr>\n  <td class="'+parity+'1">' + \
                      option.strip() + \
                      self._other_field(option, TEXTAREA) + '</td>\n'
            checkbox_index = 0
            for value in self.get(QLIKERT):
                string += '  <td class="'+parity+'"><input type="'+type+'" name="' + \
                          self.get_name(checkbox_index, likert_index) + \
                          '" value="' + form_escape(value) + \
                          '"></a></td>\n'
                checkbox_index += 1
            string += '</tr>\n'
            likert_index += 1
        string += '</table>\n\n'
        return string

    def __str__(self):
        '''Return the question as an HTML-formatted string.
        '''
        string = ''
        if self.get_type() in numbered_questions:
            string += self._question_beginning()
        if self.get_type() == V_RADIO:
            string += self._vertical_element('radio')
        elif self.get_type() == V_CHECKBOX:
            string += self._vertical_element('checkbox')
        elif self.get_type() == H_RADIO:
            string += self._horizontal_element('radio')
        elif self.get_type() == H_CHECKBOX:
            string += self._horizontal_element('checkbox')
        elif self.get_type() == SELECT:
            string += '<br><select name="' + self.get_name() + '">\n'
            for option in self.get(QOPTIONS):
                string += '  <option>' + option + '\n'
            string += '</select>'
        elif self.get_type() == TEXTINPUT:
            width = ''
            if self.has(QWIDTH):
                width = ' size="'+str(self.get(QWIDTH))+'"'
            string += '<br><input type="text" name="' + self.get_name() + '"' + width + '>\n'
        elif self.get_type() == TEXTAREA:
            width = ''
            if self.has(QWIDTH):
                width = ' cols="'+str(self.get(QWIDTH))+'"'
            height = ''
            if self.has(QHEIGHT):
                height = ' rows="'+str(self.get(QHEIGHT))+'"'
            string += '<br><textarea name="' + self.get_name() + '"' + width + height + '></textarea>\n'
        elif self.get_type() == PASSWORD:
            width = ''
            if self.has(QWIDTH):
                width = ' size="'+str(self.get(QWIDTH))+'"'
            string += '<br><input type="password" name="' + self.get_name() + '"' + width + '>\n'
        elif self.get_type() == PAGEBREAK:
            string += '<p>\n<hr>\n<p>\n'
        elif self.get_type() == PAGETITLE:
            string += '<h3>' + self.get(QTEXT) + '</h3>\n\n'
        elif self.get_type() == TEXT:
            string += '<p>\n' + self.get(QTEXT) + '<p>\n'
        elif self.get_type() == RESET:
            string += '<input type="reset">\n\n'
        else:
            pass # TODO error? unrecognized question type
        return string

    def has(self, attribute):
        '''If the question has the attribute, return true.
           Else, return false.
        '''
        return attribute in self.atts

    def get(self, attribute, alternate=None):
        '''Return the value for the attribute specified.
           (Errors for unknown attributes are not caught.)
        '''
        if alternate:
            return self.atts.get(attribute, alternate)
        return self.atts[attribute]

    def get_type(self):
        '''Return the type of the question.
        '''
        return self.get(QTYPE)

    def set(self, attribute, value):
        '''Set the specified attribute to the given value.
        '''
        self.atts[attribute] = value


class survey:
    atts = {}
    
    def __init__(self, dict={}, survey_id=''):
        '''Initialize the survey from a new survey dict
           or load a saved survey from a survey ID.
        '''        
        if not dict:
            if not survey_id:
                return # TODO error - no survey to initialize
            else:
                self.load(survey_id)
                return

        for att in survey_atts:
            self.set(att, dict[att])

        # start getting questions and pages
        question_list = []
        page_list = [0]
        
        element_index = 1
        question_number = 1
        q_code = get_code(element_index)
        while get_key(q_code, QTYPE) in dict:
            q_dict = {}
            q_dict[QCODE] = q_code
            for att in q_atts_single:
                q_dict[att] = get_value(dict, q_code, att, 1)
            if q_dict[QTYPE] in numbered_questions:
                q_dict[QNUMBER] = question_number
                question_number += 1
            elif q_dict[QTYPE] == PAGEBREAK:
                q_dict[QNUMBER] = question_number
            for att in q_atts_multi:
                values = []
                if get_value(dict, q_code, att, 1):
                    values = get_value(dict, q_code, att)
                else:
                    index = 1
                    while (get_key(q_code, likert_name(att, index))) in dict:
                        values.append(get_value(dict, q_code, likert_name(att, index)))
                        index += 1
                q_dict[att] = values
            question_list.append(question(q_dict))
            if question_list[-1].get_type() == PAGEBREAK:
                page_list.append(len(question_list))
            element_index += 1
            q_code = get_code(element_index)
            
        if not question_list[-1].get_type() == PAGEBREAK:
            question_list.append(question({QTYPE: PAGEBREAK, QCODE:q_code, QNUMBER: question_number}))
        page_list.append(len(question_list))
        self.set(QUESTIONS, question_list)
        self.set(PAGES, page_list)
        self.set(INPUT_METHOD, dict.get(INPUT_METHOD, WEB_INPUT))
        if self.get(INPUT_METHOD) == TEXT_INPUT:
            self.set(RAWTEXT, dict[RAWTEXT])
        # finished getting questions and pages

        self.set(RAWDATA, dict)

    def __str__(self):
        '''Return the whole survey as an HTML-formatted string.
        '''
        string = self.get_header()
        for page in range(1, len(self.get(PAGES))):
            string += self.get_title() + '<ol start="' + self._question_number(page) + \
                      '">\n' + self.get_page(page) + '</ol>\n<p><hr><p>\n'

        string += self.get_footer()
        return string

    def print_page(self, page, u_id='', testing=''):
        '''Return an HTML-formatted string of the specified
           survey page.
        '''
        if not u_id:
            u_id = make_id(self.get(SURVEY_ID))
        append = ''
        if testing:
            append += '\n<input type="hidden" name="' + TESTING + '" value="yes">\n'
        elif self.get_openness() == CLOSED:
            # 'elif' allows testing when closed
            return self.get_header() + CLOSED_TEXT + self.get_footer()
        if not self.last_page(page):
            append += '\n<input type="submit" value="Continue">\n'
        string = self.get_header() + self.get_title() + \
                 '<form method="post" action="' + PROCESS_SURVEY_CGI + '">\n' + \
                 '<input type="hidden" name="' + SURVEY_ID + '" value="' + \
                 self.get(SURVEY_ID) + '">\n' + \
                 '<input type="hidden" name="'+USER_ID+'" value="' + \
                 u_id + '">\n' + \
                '<input type="hidden" name="'+PAGENUM+'" value="' + \
                 str(page)+'">\n' + \
                 '<ol start="' + self._question_number(page) + '">\n\n' + \
                 self.get_page(page, u_id) + \
                 '</ol>' + append + '</form>\n' + \
                 '<p><hr><p>' + \
                 self.get_footer()
        return string

    def get_modifiers(self):
        '''Return the modifications dict of the survey.
        '''
        s_id = self.get(SURVEY_ID)
        try:
            file = open(get_survey_dir(s_id, 0)+SURVEY_MODS)
            modifiers = pickle.load(file)
            file.close()
        except:
            modifiers = {}
        return modifiers

    def save_modifiers(self, modifiers):
        '''Save the modifications dict of the survey in
           SURVEY_MODS file.
        '''
        s_id = self.get(SURVEY_ID)
        file = open(get_survey_dir(s_id, 0)+SURVEY_MODS, 'w')
        pickle.dump(modifiers, file)
        file.close()

    def get_openness(self, give_default=1):
        '''Return "'''+OPEN+'''" if the survey is open,
           "'''+CLOSED+'''" otherwise.  If openness
           isn't set, return the default openness (unless
           give_default is 0).
        '''
        openness = ''
        if give_default:
            openness = DEFAULT_OPENNESS
        modifiers = self.get_modifiers()
        opendate = modifiers.get(OPEN_DATE, '')
        closedate = modifiers.get(CLOSE_DATE, '')
        current = time.time()
        if opendate:
            if current < opendate: # before opening
                openness = CLOSED
            else: # after opening
                # after opening AND closing case handled below
                openness = OPEN
        if closedate:
            if current > closedate: # after closing
                openness = CLOSED
            elif opendate and current >= opendate:
                # between opening and closing
                openness = OPEN
        return openness
    
    def set_openness(self, year, month, day, hours, minutes, attribute):
        '''Set OPEN_DATE or CLOSE_DATE for survey.  Give all time values as
           0 to remove. ("attribute" must be OPEN_DATE or CLOSE_DATE.)
        '''
        modifiers = self.get_modifiers()
        if not (year+month+day+hours+minutes):
            modifiers[attribute] = ''
        else:
            modifiers[attribute] = time.mktime((year, month, day, hours, minutes, 0, 0, 0, 0))
        self.save_modifiers(modifiers)

    def last_page(self, page):
        '''Return true if the page given is the last page
           of the survey.
        '''
        return page >= (len(self.get(PAGES))-1)
    
    def get(self, attribute, alternate=None):
        '''Return the specified attribute of the survey.
           [Errors are not caught.]
        '''
        if not alternate == None:
            return self.atts.get(attribute, alternate)
        else:
            return self.atts[attribute]

    def set(self, attribute, value):
        '''Set the specified attribute of the survey to
           the value given.
        '''
        if attribute in self.atts:
            self.atts[attribute] = value
        else:
            self.atts[attribute] = value

    def save(self, u_id):
        '''Save the survey in its directory, overwriting
           any previous saves.  Create survey directory if it
           doesn't exist.
        '''
        s_id = self.get(SURVEY_ID)
        if s_id not in listdir(SURVEY_PATH):
            os.makedirs(get_survey_dir(s_id))
        secure_survey(s_id, u_id)
        file = open(get_survey_dir(s_id, 0)+SURVEY_FILE, 'w')
        pickle.dump(self.atts, file)
        file.close()

    def load(self, survey_id):
        '''Load the survey specified by survey_id from its
           directory.
        '''
        file = open(get_survey_dir(survey_id, 0, 0)+SURVEY_FILE)
        self.atts = pickle.load(file)
        file.close()

    def get_header(self):
        '''Return the survey's header.
        '''
        return get_header(self.get(SURVEY_ID))

    def get_title(self):
        '''Return the survey's title, HTML-formatted.
        '''
        return '<h1>' + self.get(TITLE) + '</h1>\n\n'
    
    def get_footer(self):
        '''Return the survey's footer.
        '''
        return get_footer(self.get(SURVEY_ID))

    def _question_number(self, page):
        if page == 1:
            return '1'
        else:
            return str(self.get(QUESTIONS)[self.get(PAGES)[page-1]-1].get(QNUMBER))

    def code_check(self, string, USER_ID='', PAGENUM=''):
        for word in CODEWORDS:
            string = string.replace(word, eval(CODEWORDS[word]))
        return string                                    

    def get_page(self, page, u_id=''):
        '''Return the questions in the given page as an
           HTML-formatted string.
        '''
    	start = self.get(PAGES)[page-1]
	end = self.get(PAGES)[page]-1	# the -1 excludes pagebreaks
        questions = self.get(QUESTIONS)[start:end]
        string = ''
        for question in questions:
            string += question.__str__()
        return self.code_check(string, u_id, page)

    def save_results(self, dict):
        '''Save the survey results dictionary in a unique file of
           <id>_<pagenum> in the survey's data directory using
           the pickle module.
        '''
        dict[TIMESTAMP] = time.ctime()
        pagenum = get_pagenum(dict)
        user_id = get_survey_value(dict, USER_ID)
        file = open(get_survey_dir(self.get(SURVEY_ID))+user_id+'_'+str(pagenum), 'w')
        pickle.dump(dict, file)
        file.close()

    def process(self, dict, testing=''):
        '''If pagenum is in dict, save current page results and
           print the next page. Else, print first page of survey.
        '''
        if PAGENUM in dict:
            if not testing:
                self.save_results(dict)
            print self.print_page(get_pagenum(dict)+1, get_survey_value(dict, USER_ID), testing=testing)
        else:
            print self.print_page(1, testing=testing)

    # RESULTS PROCESSING

    def get_question_text(self, include):
        '''Return a list of the names of all the questions
           in the survey (not the question text), or the
           subset specified by include.
        '''
        questions = {}
        included_questions = numbered_questions
        if include == INCLUDE_MULTI:
            included_questions = multiple_choice_qs
        elif include == INCLUDE_TEXT:
            included_questions = text_questions
        for question in self.get(QUESTIONS):
            if question.get_type() in included_questions:
                questions[question.get_name()] = question.get(QTEXT)
                # TODO options?
        return questions

    def output_text(self, include):
        question_dict = self.get_question_text(include)
        names = self.get_question_list(include, expand_multi=0)
        text = []
        for name in names:
            text.append(csv_escape(question_dict.get(name, '')))
        return ','.join(names) + '\n"' + '","'.join(text) + '"'

    def get_question_list(self, include, expand_multi=1):
        '''Return a list of the names of all the questions
           in the survey (not the question text), or the
           subset specified by include.
        '''
        questions = []
        included_questions = numbered_questions
        if include == INCLUDE_MULTI:
            included_questions = multiple_choice_qs
        elif include == INCLUDE_TEXT:
            included_questions = text_questions
        for question in self.get(QUESTIONS):
            if question.get_type() in included_questions:
                if expand_multi:
                    questions.extend(question.get_names())
                else:
                    questions.append(question.get_name())
        return questions
        '''
        # OLD WAY OF DOING IT - JUST GRABBING DICT
        # NAMES.  REWRITTEN TO ACTUALLY LOAD NAMES
        # FROM SURVEY.
        questions = []
        survey_dir = get_survey_dir(self.get(SURVEY_ID))
        files = os.listdir(survey_dir)
        ids = {}
        for file in files:
            (id, page) = file.split('_')
            dict = ids.get(id, {})
            filehandle = open(survey_dir + file)
            page_dict = pickle.load(filehandle)
            filehandle.close()
            for key in page_dict.keys():
                if key not in (USER_ID, PAGENUM) and \
                   key not in questions:
                    questions.append(key)
        questions.sort()
        if SURVEY_ID in questions:
            questions.remove(SURVEY_ID)
        return questions
        '''

    def get_result_ids(self):
        '''Return a list of the ids of all respondents.
        '''
        survey_dir = get_survey_dir(self.get(SURVEY_ID))
        files = os.listdir(survey_dir)
        ids = []
        for file in files:
            try:
                id = file.split('_')[0]
                if id not in ids:
                    ids.append(id)
            except:
                continue
        return ids
            
    def get_results_dict(self):
        '''Make a dictionary, keyed by respondent ID, of dictionaries
           of responses keyed by input field name for the survey
           specified by survey_id.
        '''
        survey_dir = get_survey_dir(self.get(SURVEY_ID))
        files = os.listdir(survey_dir)
        ids = {}
        for file in files:
            try:
                (id, page) = file.split('_')
            except:
                continue
            dict = ids.get(id, {})
            filehandle = open(survey_dir + file)
            page_dict = pickle.load(filehandle)
            filehandle.close()
            for key in page_dict.keys():
                if key not in (USER_ID, PAGENUM):
                    dict[key] = get_survey_value(page_dict, key)
            ids[id] = dict
        return ids

    def output_csv(self, include):
        '''Output data in data directory as comma-separated value
           (CSV) text with a header line and with all values
           surrounded by double-quotes for the survey specified
           by survey_id.
        '''
        results_dict = self.get_results_dict()
        question_list = self.get_question_list(include)
        output = 'id,' + ','.join(question_list) + '\n'
        for id in results_dict:
            questions = [id]
            for question in question_list:
                answer = csv_escape(results_dict[id].get(question, ''))
                if "'" in answer or ',' in answer:
                    answer = '"' + answer + '"'
                questions.append(answer)
            output += ','.join(questions) + '\n'
        return output


def get_results(dict):
    '''Print the results specified.
    '''
    print 'Content-Type: text/plain\n'
    if dict[FORMAT] == QUESTIONS_ONLY:
        print survey(survey_id=dict[SURVEY_ID]).output_text(dict[INCLUDE])
    elif dict[FORMAT] == CSV_TEXT:
        print survey(survey_id=dict[SURVEY_ID]).output_csv(dict[INCLUDE])
    elif dict[FORMAT] == CSV_NUM:
        pass # TODO
    elif dict[FORMAT] == SPSS_DATA:
        pass # TODO
    elif dict[FORMAT] == SPSS_VARS:
        pass # TODO
    elif dict[FORMAT] == SPSS_VARS:
        pass # TODO
    elif dict[FORMAT] == HTML_FORMAT:
        pass # TODO
